<?php
/**
 * Version 1.0
 * Date: 6/21/2018
 *
 * SiteGuarding.com Antivirus Registration
 * 
 * Usage: /usr/bin/php registration.php email
 */

//Init
ini_set('register_argc_argv','1'); 

if (PHP_SAPI !== 'cli') {
    echo 'ERROR: registration.php must be run as a CLI application';
    exit(1);
}

if(!extension_loaded('curl')) {
    echo 'ERROR: curl must be installed and configured as an extension for php';
    exit(1);
}

if(!ini_get('register_argc_argv')) {
    echo 'ERROR: register_argc_argv must be true';
    exit(1);
}

if (count($argv) < 2) {
    echo "ERROR: Invalid number of arguments\n";
    echo "Usage: /usr/bin/php registration.php email";
    exit(1);
}

if(!filter_var($argv[1], FILTER_VALIDATE_EMAIL)) {
    echo "ERROR: Invalid email format";
    exit(1);
}

define('SITEGUARDING_SERVER_HTTP', 'http://www.siteguarding.com/ext/antivirus/index.php');
define('SITEGUARDING_SERVER_HTTPS', 'https://www.siteguarding.com/ext/antivirus/index.php');

$link = SITEGUARDING_SERVER_HTTP.'?action=register&type=json&data=';
$link_https = SITEGUARDING_SERVER_HTTPS.'?action=register&type=json&data=';

$domain = php_uname("n");
$email = trim($argv[1]);
$access_key = md5(time().$domain.rand(1, 20000).$email);

$data = array(
	'domain' => $domain,
	'email' => $email,
	'access_key' => $access_key,
    'cms' => 'server',
);

$link .= base64_encode(json_encode($data));
$link_https .= base64_encode(json_encode($data));

$response = trim(GetRemote_file_contents($link));

if($response === false) {
    $response = trim(GetRemote_file_contents($link_https));
    if($response === false) {
        echo "ERROR: Your server cant communicate with Siteguarding server. Please contact to Siteguarding support team.";
        exit(1);
    }
}

if($response === '') {
    create_config($domain,$access_key);
    echo "registration has been successfully completed";
    unlink('sg_antivirus_scanner.conf');
    unlink('sg_av_logo.png');
    unlink('registration.php');
} elseif(stripos($response,'is not exist') !== false) {
    echo "ERROR: The domain is not exist or couldnt be resolved. Registration is available only for real domain. Please contact our support by email support@siteguarding.com";
    exit(1);
} elseif(stripos($response,'is already registered') !== false) {
    echo "ERROR: Domain is already registered. Please contact to Siteguarding support team.";
    exit(1);
}

function GetRemote_file_contents($url, $parse = false) {
    
    $ch = curl_init();
            
    curl_setopt($ch, CURLOPT_URL, $url );
    curl_setopt($ch, CURLOPT_USERAGENT, "Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0)");
    curl_setopt($ch, CURLOPT_TIMEOUT, 3600);
    curl_setopt($ch, CURLOPT_TIMEOUT_MS, 3600000);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);

    curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 10); 
    curl_setopt($ch, CURLOPT_CONNECTTIMEOUT_MS, 10000); 
    curl_setopt($ch, CURLOPT_BINARYTRANSFER, 1);
    curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
            
    $output = trim(curl_exec($ch));
    curl_close($ch);
            
    if ($output === false) {
        return false;
    } 
    if(trim($output) == '') {
       return ''; 
    }       
    if ($parse === true) {
        $output = (array)json_decode($output, true);
    } 
            
    return $output;
}

function create_config($domain,$access_key) {
    
    $data = '<?php
    define("DOMAIN","'.$domain.'");
    define("ACCESS_KEY","'.$access_key.'");
    define("ANTIVIRUS_PLATFORM","server");
    define("SCRIPT_VERSION","1.0");
    define("SITEGUARDING_SERVER","'.SITEGUARDING_SERVER_HTTPS.'");';
    
    $fp = fopen('config.php','w');
    if($fp === false) {
        echo "ERROR: Cant create config.php file. Please contact to Siteguarding support team.";
        exit(1);
    }
    fwrite($fp,$data);
    fclose($fp);
    chmod('config.php', 0750);
}



